/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.gan;

import static net.brutex.ai.dnn.api.NN.*;

import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.*;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.impl.ActivationLReLU;
import org.nd4j.linalg.learning.config.Adam;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.lossfunctions.LossFunctions;

public class App2Config {

  public static final int INPUT = 100;
  public static final int BATCHSIZE=150;
  public static final int X_DIM = 28;
  public static final int Y_DIM = 28;
  public static final int CHANNELS = 1;
  public static final int EPOCHS = 50;
  public static final IUpdater UPDATER = Adam.builder().learningRate(0.0002).beta1(0.5).build();
  public static final IUpdater UPDATER_DIS = Adam.builder().learningRate(0.02).beta1(0.5).build();
  public static final boolean SHOW_GENERATED = true;
  public static final float COLORSPACE = 255f;

  final static int OUTPUT_PER_PANEL = 10;

  static LayerConfiguration[] genLayerConfig() {
    return new LayerConfiguration[] {
            /*
      DenseLayer.builder().name("L-0").nIn(INPUT).nOut(INPUT + (INPUT / 2)).activation(Activation.RELU).build(),
      ActivationLayer.builder().activation(Activation.RELU).build(), /*
                Deconvolution2D.builder().name("L-Deconv-01").nIn(CHANNELS).nOut(CHANNELS)
                        .kernelSize(2,2)
                        .stride(1,1)
                        .padding(0,0)
                        .convolutionMode(ConvolutionMode.Truncate)
                        .activation(Activation.RELU)
                        .hasBias(BIAS).build(),
                //BatchNormalization.builder().nOut(CHANNELS).build(),
                Deconvolution2D.builder().name("L-Deconv-02").nIn(CHANNELS).nOut(CHANNELS)
                        .kernelSize(2,2)
                        .stride(2,2)
                        .padding(0,0)
                        .convolutionMode(ConvolutionMode.Truncate)
                        .activation(Activation.RELU)
                        .hasBias(BIAS).build(),
               //BatchNormalization.builder().name("L-batch").nOut(CHANNELS).build(),


      DenseLayer.builder().name("L-x").nIn(INPUT + (INPUT / 2)).nOut(2 * INPUT).build(),
      ActivationLayer.builder().activation(Activation.RELU).dropOut(0.2).build(),
      DenseLayer.builder().name("L-x").nIn(2 * INPUT).nOut(3 * INPUT).build(),
      ActivationLayer.builder().activation(Activation.RELU).dropOut(0.2).build(),
      DenseLayer.builder().name("L-x").nIn(3 * INPUT).nOut(2 * INPUT).build(),
      ActivationLayer.builder().activation(Activation.RELU).dropOut(0.2).build(),
      // DropoutLayer.builder(0.001).build(),
      DenseLayer.builder().nIn(2 * INPUT).nOut(INPUT).activation(Activation.TANH).build()    */

            dense().nIn(INPUT).nOut(256).weightInit(WeightInit.NORMAL).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            dense().nIn(256).nOut(512).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            dense().nIn(512).nOut(1024).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            dense().nIn(1024).nOut(784).activation(Activation.TANH).build(),

    };
  }

  static LayerConfiguration[] disLayerConfig() {
    return new LayerConfiguration[] {/*
                Convolution2D.builder().nIn(CHANNELS).kernelSize(2,2).padding(1,1).stride(1,1).nOut(CHANNELS)
                        .build(),
                Convolution2D.builder().nIn(CHANNELS).kernelSize(3,3).padding(1,1).stride(2,2).nOut(CHANNELS)
                        .build(),
                ActivationLayer.builder().activation(Activation.LEAKYRELU).build(),
                BatchNormalization.builder().build(),
                OutputLayer.builder().nOut(1).lossFunction(LossFunctions.LossFunction.MCXENT)
                        .activation(Activation.SIGMOID)
                        .build()


            dense().name("L-dense").nIn(INPUT).nOut(INPUT).build(),
            ActivationLayer.builder().activation(Activation.RELU).build(),
            DropoutLayer.builder(0.5).build(),

            DenseLayer.builder().nIn(INPUT).nOut(INPUT/2).build(),
            ActivationLayer.builder().activation(Activation.RELU).build(),
            DropoutLayer.builder(0.5).build(),

            DenseLayer.builder().nIn(INPUT/2).nOut(INPUT/4).build(),
            ActivationLayer.builder().activation(Activation.RELU).build(),
            DropoutLayer.builder(0.5).build(),

            OutputLayer.builder().nIn(INPUT/4).nOut(1).lossFunction(LossFunctions.LossFunction.XENT)
                    .activation(Activation.SIGMOID)
                    .build() */
            dense().nIn(784).nOut(1024).hasBias(true).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            DropoutLayer.builder(1 - 0.5).build(),
            dense().nIn(1024).nOut(512).hasBias(true).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            DropoutLayer.builder(1 - 0.5).build(),
            dense().nIn(512).nOut(256).hasBias(true).build(),
            ActivationLayer.builder(new ActivationLReLU(0.2)).build(),
            DropoutLayer.builder(1 - 0.5).build(),
            OutputLayer.builder(LossFunctions.LossFunction.XENT).nIn(256).nOut(1).activation(Activation.SIGMOID).build()
    };
  }


  static NeuralNetConfiguration generator() {
    NeuralNetConfiguration conf =
            NeuralNetConfiguration.builder()
                    .name("generator")
                    .gradientNormalization(GradientNormalization.RenormalizeL2PerLayer)
                    .gradientNormalizationThreshold(100)
                    .seed(42)
                    .updater(UPDATER)
                    .weightInit(WeightInit.XAVIER)
                    //.weightNoise(new WeightNoise(new NormalDistribution(0.5, 0.5)))
                    .weightNoise(null)
                    // .weightInitFn(new WeightInitXavier())
                    // .activationFn(new ActivationIdentity())
                    .activation(Activation.IDENTITY)
                    .layersFromArray(App2Config.genLayerConfig())
                    // .inputType(InputType.convolutional(DIMENSIONS, DIMENSIONS, CHANNELS))
                    //.inputPreProcessor(0, new CnnToFeedForwardPreProcessor())
                    //.inputPreProcessor(2, new FeedForwardToCnnPreProcessor(DIMENSIONS, DIMENSIONS, CHANNELS))
                    //.inputPreProcessor(4, new CnnToFeedForwardPreProcessor())

                    .build();
    conf.init();
    return conf;
  }

  static NeuralNetConfiguration discriminator() {
    NeuralNetConfiguration conf =
            NeuralNetConfiguration.builder()
                    .name("discriminator")
                    .gradientNormalization(GradientNormalization.RenormalizeL2PerLayer)
                    .gradientNormalizationThreshold(100)
                    .seed(42)
                    .updater(UPDATER_DIS)
                    .weightInit(WeightInit.XAVIER)
                    // .weightNoise(new WeightNoise(new NormalDistribution(0.5, 0.5)))
                    .weightNoise(null)
                    // .weightInitFn(new WeightInitXavier())
                    // .activationFn(new ActivationIdentity())
                    .activation(Activation.IDENTITY)
                    .layersFromArray(disLayerConfig())
                    //.inputPreProcessor(0, new FeedForwardToCnnPreProcessor(DIMENSIONS, DIMENSIONS, CHANNELS))
                    //.inputPreProcessor(0, new CnnToFeedForwardPreProcessor())
                    //.dataType(DataType.FLOAT)
                    .build();
    conf.init();
    return conf;
  }
}
